import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
from scipy import linalg
from scipy.special import logsumexp
from sklearn.mixture import GaussianMixture
from sklearn.cluster import KMeans
from sklearn.utils import check_array
from sklearn.utils.validation import check_is_fitted
from sklearn.mixture._gaussian_mixture import (
    _compute_precision_cholesky,
    _compute_log_det_cholesky,
    _estimate_gaussian_covariances_full,
    _estimate_gaussian_covariances_diag,
    _estimate_gaussian_covariances_spherical,
    _check_means,
    _check_precisions,
    _check_shape,
)
from sklearn.mixture._base import check_random_state, ConvergenceWarning
import warnings
from SimPEG.maps import IdentityMap


###############################################################################
# Disclaimer: the following classes built upon the GaussianMixture class      #
# from Scikit-Learn. New functionalitie are added, as well as modifications to#
# existing functions, to serve the purposes pursued within SimPEG.            #
# This use is allowed by the Scikit-Learn licensing (BSD-3-Clause License)    #
# and we are grateful for their contributions to the open-source community.   #                                                   #
###############################################################################


class WeightedGaussianMixture(GaussianMixture):
    """
    Weighted Gaussian mixture class

    This class upon the GaussianMixture class from Scikit-Learn.
    Two main modifications:

        1. Each sample/observation is given a weight, the volume of the
        corresponding discretize.BaseMesh cell, when fitting the
        Gaussian Mixture Model (GMM). More volume gives more importance, ensuing a
        mesh-free evaluation of the clusters of the geophysical model.

        2. When set manually, the proportions can be set either globally (normal behavior)
        or cell-by-cell (improvements)

    Disclaimer: this class built upon the GaussianMixture class from Scikit-Learn.
    New functionalitie are added, as well as modifications to
    existing functions, to serve the purposes pursued within SimPEG.
    This use is allowed by the Scikit-Learn licensing (BSD-3-Clause License)
    and we are grateful for their contributions to the open-source community.

    Addtional parameters to provide, compared to sklearn.mixture.gaussian_mixture:

    Parameters
    ----------
    n_components : int
        Number of components
    mesh : discretize.BaseMesh
        ``TensorMesh`` or ``QuadTree`` or Octree) mesh: the volume
        of the cells give each sample/observations its weight in the fitting proces
    actv : numpy.ndarry, optional
        Active indexes
    """

    def __init__(
        self,
        n_components,
        mesh,
        actv=None,
        covariance_type="full",
        init_params="kmeans",
        max_iter=100,
        means_init=None,
        n_init=10,
        precisions_init=None,
        random_state=None,
        reg_covar=1e-06,
        tol=0.001,
        verbose=0,
        verbose_interval=10,
        warm_start=False,
        weights_init=None,
        # **kwargs
    ):
        self.mesh = mesh
        self.actv = actv
        if self.actv is None:
            self.cell_volumes = self.mesh.cell_volumes
        else:
            self.cell_volumes = self.mesh.cell_volumes[self.actv]

        super(WeightedGaussianMixture, self).__init__(
            covariance_type=covariance_type,
            init_params=init_params,
            max_iter=max_iter,
            means_init=means_init,
            n_components=n_components,
            n_init=n_init,
            precisions_init=precisions_init,
            random_state=random_state,
            reg_covar=reg_covar,
            tol=tol,
            verbose=verbose,
            verbose_interval=verbose_interval,
            warm_start=warm_start,
            weights_init=weights_init,
            # **kwargs
        )
        # set_kwargs(self, **kwargs)

    def compute_clusters_precisions(self):
        """Compute and set the precisions matrices and their Cholesky decomposition.

        Use this function after setting covariances manually.
        """
        self.precisions_cholesky_ = _compute_precision_cholesky(
            self.covariances_, self.covariance_type
        )
        if self.covariance_type == "full":
            self.precisions_ = np.empty(self.precisions_cholesky_.shape)
            for k, prec_chol in enumerate(self.precisions_cholesky_):
                self.precisions_[k] = np.dot(prec_chol, prec_chol.T)

        elif self.covariance_type == "tied":
            self.precisions_ = np.dot(
                self.precisions_cholesky_, self.precisions_cholesky_.T
            )
        else:
            self.precisions_ = self.precisions_cholesky_**2

    def compute_clusters_covariances(self):
        """Compute the precisions matrices and their Cholesky decomposition.

        Use this function after setting precisions matrices manually.
        """
        self.covariances_cholesky_ = _compute_precision_cholesky(
            self.precisions_, self.covariance_type
        )
        if self.covariance_type == "full":
            self.covariances_ = np.empty(self.covariances_cholesky_.shape)
            for k, cov_chol in enumerate(self.covariances_cholesky_):
                self.covariances_[k] = np.dot(cov_chol, cov_chol.T)

        elif self.covariance_type == "tied":
            self.covariances_ = np.dot(
                self.covariances_cholesky_, self.covariances_cholesky_.T
            )
        else:
            self.covariances_ = self.covariances_cholesky_**2

        self.precisions_cholesky_ = _compute_precision_cholesky(
            self.covariances_, self.covariance_type
        )

    def order_clusters_GM_weight(self, outputindex=False):
        """Order clusters by decreasing weights

        PARAMETERS
        ----------
        outputindex : bool, default: ``True``
            If ``True``, return the sorting index

        RETURN
        ------
        np.ndarray
            Sorting index
        """
        if self.weights_.ndim == 1:
            indx = np.argsort(self.weights_, axis=0)[::-1]
            self.weights_ = self.weights_[indx].reshape(self.weights_.shape)
        else:
            indx = np.argsort(self.weights_.sum(axis=0), axis=0)[::-1]
            self.weights_ = self.weights_[:, indx].reshape(self.weights_.shape)

        self.means_ = self.means_[indx].reshape(self.means_.shape)

        if self.covariance_type == "tied":
            pass
        else:
            self.precisions_ = self.precisions_[indx].reshape(self.precisions_.shape)
            self.covariances_ = self.covariances_[indx].reshape(self.covariances_.shape)

        self.precisions_cholesky_ = _compute_precision_cholesky(
            self.covariances_, self.covariance_type
        )

        if outputindex:
            return indx

    def _check_weights(self, weights, n_components, n_samples):
        """
        [modified from Scikit-Learn.mixture.gaussian_mixture]
        Check the user provided 'weights'.
        Parameters
        ----------
        weights : array-like, shape (n_components,) or (n_samples, n_components)
            The proportions of components of each mixture.
        n_components : int
            Number of components.

        Returns
        -------
        weights : array, shape (n_components,)
        """

        if len(weights.shape) == 2:
            weights = check_array(
                weights, dtype=[np.float64, np.float32], ensure_2d=True
            )
            _check_shape(weights, (n_samples, n_components), "weights")
        else:
            weights = check_array(
                weights, dtype=[np.float64, np.float32], ensure_2d=False
            )
            _check_shape(weights, (n_components,), "weights")

        # check range
        if any(np.less(weights, 0.0)) or any(np.greater(weights, 1.0)):
            raise ValueError(
                "The parameter 'weights' should be in the range "
                "[0, 1], but got max value %.5f, min value %.5f"
                % (np.min(weights), np.max(weights))
            )

        # check normalization
        if not np.allclose(np.abs(1.0 - np.sum(weights.T, axis=0)), 0.0):
            raise ValueError(
                "The parameter 'weights' should be normalized, "
                "but got sum(weights) = %.5f" % np.sum(weights)
            )

        return weights

    def _check_parameters(self, X):
        """
        [modified from Scikit-Learn.mixture.gaussian_mixture]
        Check the Gaussian mixture parameters are well defined.
        """
        n_samples, n_features = X.shape
        if self.covariance_type not in ["spherical", "tied", "diag", "full"]:
            raise ValueError(
                "Invalid value for 'covariance_type': %s "
                "'covariance_type' should be in "
                "['spherical', 'tied', 'diag', 'full']" % self.covariance_type
            )

        if self.weights_init is not None:
            self.weights_init = self._check_weights(
                self.weights_init,
                self.n_components,
                n_samples,
            )

        if self.means_init is not None:
            self.means_init = _check_means(
                self.means_init, self.n_components, n_features
            )

        if self.precisions_init is not None:
            self.precisions_init = _check_precisions(
                self.precisions_init,
                self.covariance_type,
                self.n_components,
                n_features,
            )

    def _initialize_parameters(self, X, random_state):
        """
        [modified from Scikit-Learn.mixture._base]
        Initialize the model parameters.
        Parameters
        ----------
        X : array-like, shape  (n_samples, n_features)
        random_state : RandomState
            A random number generator instance.
        """
        n_samples, _ = X.shape

        if self.init_params == "kmeans":
            resp = np.zeros((n_samples, self.n_components))
            label = (
                KMeans(
                    n_clusters=self.n_components, n_init=1, random_state=random_state
                )
                .fit(X, sample_weight=self.cell_volumes)
                .labels_
            )
            resp[np.arange(n_samples), label] = 1
        elif self.init_params == "random":
            resp = random_state.rand(n_samples, self.n_components)
            resp /= resp.sum(axis=1)[:, np.newaxis]
        else:
            raise ValueError(
                "Unimplemented initialization method '%s'" % self.init_params
            )

        self._initialize(X, resp)

    def _m_step(self, X, log_resp):
        """
        [modified from Scikit-Learn.mixture.gaussian_mixture]
        M step.
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
        log_resp : array-like, shape (n_samples, n_components)
            Logarithm of the posterior probabilities (or responsibilities) of
            the point of each sample in X.
        """
        n_samples, _ = X.shape
        Volume = np.mean(self.cell_volumes)
        weights, self.means_, self.covariances_ = self._estimate_gaussian_parameters(
            X, self.mesh, np.exp(log_resp), self.reg_covar, self.covariance_type
        )
        weights /= n_samples * Volume
        self.precisions_cholesky_ = _compute_precision_cholesky(
            self.covariances_, self.covariance_type
        )

        if len(self.weights_.shape) == 1:
            self.weights_ = weights

    def _estimate_gaussian_covariances_tied(self, resp, X, nk, means, reg_covar):
        """
        [modified from Scikit-Learn.mixture.gaussian_mixture]
        Estimate the tied covariance matrix.
        Parameters
        ----------
        resp : array-like, shape (n_samples, n_components)
        X : array-like, shape (n_samples, n_features)
        nk : array-like, shape (n_components,)
        means : array-like, shape (n_components, n_features)
        reg_covar : float
        Returns
        -------
        covariance : array, shape (n_features, n_features)
        The tied covariance matrix of the components.
        """
        avg_X2 = np.dot(self.cell_volumes * X.T, X)
        avg_means2 = np.dot(nk * means.T, means)
        covariance = avg_X2 - avg_means2
        covariance /= nk.sum()
        covariance.flat[:: len(covariance) + 1] += reg_covar
        return covariance

    def _estimate_gaussian_parameters(self, X, mesh, resp, reg_covar, covariance_type):
        """
        [modified from Scikit-Learn.mixture.gaussian_mixture]
        Estimate the Gaussian distribution parameters.
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input data array.
        resp : array-like, shape (n_samples, n_components)
            The responsibilities for each data sample in X.
        reg_covar : float
            The regularization added to the diagonal of the covariance matrices.
        covariance_type : {'full', 'tied', 'diag', 'spherical'}
            The type of precision matrices.
        Returns
        -------
        nk : array-like, shape (n_components,)
            The numbers of data samples in the current components.
        means : array-like, shape (n_components, n_features)
            The centers of the current components.
        covariances : array-like
            The covariance matrix of the current components.
            The shape depends of the covariance_type.
        """
        respVol = self.cell_volumes.reshape(-1, 1) * resp
        nk = respVol.sum(axis=0) + 10 * np.finfo(resp.dtype).eps
        means = np.dot(respVol.T, X) / nk[:, np.newaxis]
        covariances = {
            "full": _estimate_gaussian_covariances_full,
            "tied": self._estimate_gaussian_covariances_tied,
            "diag": _estimate_gaussian_covariances_diag,
            "spherical": _estimate_gaussian_covariances_spherical,
        }[covariance_type](respVol, X, nk, means, reg_covar)
        return nk, means, covariances

    def _e_step(self, X):
        """
        [modified from Scikit-Learn.mixture.gaussian_mixture]
        E step.
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
        Returns
        -------
        log_prob_norm : float
            Mean of the logarithms of the probabilities of each sample in X
        log_responsibility : array, shape (n_samples, n_components)
            Logarithm of the posterior probabilities (or responsibilities) of
            the point of each sample in X.
        """
        log_prob_norm, log_resp = self._estimate_log_prob_resp(X)
        return np.average(log_prob_norm, weights=self.cell_volumes), log_resp

    def score(self, X, y=None):
        """Compute the per-sample average log-likelihood

        [modified from Scikit-Learn.mixture.gaussian_mixture]
        Compute the per-sample average log-likelihood of the given data X.

        Parameters
        ----------
        X : (n_samples, n_dimensions) array-like
            List of n_features-dimensional data points. Each row
            corresponds to a single data point.
        y : ``None``
            Placeholder variable

        Returns
        -------
        float
            Log likelihood of the Gaussian mixture given X.
        """
        return np.average(self.score_samples(X), weights=self.cell_volumes)

    def _estimate_log_gaussian_prob_with_sensW(
        self, X, sensW, means, precisions_chol, covariance_type
    ):
        """
        [New function, modified from Scikit-Learn.mixture.gaussian_mixture._estimate_log_gaussian_prob]
        Estimate the log Gaussian probability with depth or sensitivity weighting.
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
        means : array-like, shape (n_components, n_features)
        sensW: array-like, Sensitvity or Depth Weighting, shape(n_samples, n_features)
        precisions_chol : array-like,
            Cholesky decompositions of the precision matrices.
            'full' : shape of (n_components, n_features, n_features)
            'tied' : shape of (n_features, n_features)
            'diag' : shape of (n_components, n_features)
            'spherical' : shape of (n_components,)
        covariance_type : {'full', 'tied', 'diag', 'spherical'}
        Returns
        -------
        log_prob : array, shape (n_samples, n_components)
        """
        n_samples, n_features = X.shape
        n_components, _ = means.shape
        # det(precision_chol) is half of det(precision)
        log_det = _compute_log_det_cholesky(
            precisions_chol, covariance_type, n_features
        )

        if covariance_type == "full":
            log_prob = np.empty((n_samples, n_components))
            for k, (mu, prec_chol) in enumerate(zip(means, precisions_chol)):
                y = np.dot(X * sensW, prec_chol) - np.dot(mu * sensW, prec_chol)
                log_prob[:, k] = np.sum(np.square(y), axis=1)

        elif covariance_type == "tied":
            log_prob = np.empty((n_samples, n_components))
            for k, mu in enumerate(means):
                y = np.dot(X * sensW, precisions_chol) - np.dot(
                    mu * sensW, precisions_chol
                )
                log_prob[:, k] = np.sum(np.square(y), axis=1)

        else:
            log_prob = np.empty((n_samples, n_components))
            for k, (mu, prec_chol) in enumerate(zip(means, precisions_chol)):
                prec_chol_mat = np.eye(n_features) * prec_chol
                y = np.dot(X * sensW, prec_chol_mat) - np.dot(mu * sensW, prec_chol_mat)
                log_prob[:, k] = np.sum(np.square(y), axis=1)

        return -0.5 * (n_features * np.log(2 * np.pi) + log_prob) + log_det

    def _estimate_log_prob_with_sensW(self, X, sensW):
        """
        [New function, modified from Scikit-Learn.mixture.gaussian_mixture._estimate_log_prob]
        """
        return self._estimate_log_gaussian_prob_with_sensW(
            X, sensW, self.means_, self.precisions_cholesky_, self.covariance_type
        )

    def _estimate_weighted_log_prob_with_sensW(self, X, sensW):
        """
        [New function, modified from Scikit-Learn.mixture.gaussian_mixture._estimate_weighted_log_prob]
        Estimate the weighted log-probabilities, log P(X | Z) + log weights.
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
        Returns
        -------
        weighted_log_prob : array, shape (n_samples, n_component)
        """
        return (
            self._estimate_log_prob_with_sensW(X, sensW) + self._estimate_log_weights()
        )

    def score_samples_with_sensW(self, X, sensW):
        """Compute the weighted log probabilities for each sample.

        [New function, modified from Scikit-Learn.mixture.gaussian_mixture.score_samples]
        Compute the weighted log probabilities for each sample.

        Parameters
        ----------
        X : (n_samples, n_features) array_like
            List of n_features-dimensional data points. Each row
            corresponds to a single data point.
        sensW : (n_samples) array_like
            Sensitivity weights

        Returns
        -------
        (n_samples) numpy.array
            Log probabilities of each data point in X.
        """
        check_is_fitted(self)
        X = self._validate_data(X, reset=False)

        return logsumexp(self._estimate_weighted_log_prob_with_sensW(X, sensW), axis=1)

    def plot_pdf(
        self,
        ax=None,
        flag2d=False,
        x_component=None,
        y_component=None,
        padding=0.2,
        plotting_precision=100,
        plot_membership=False,
        contour_opts=None,
        level_opts=None,
    ):
        """
        Utils to plot the marginal PDFs of a GMM, either in 1D or 2D (1 or 2 physical properties at the time).

        Parameters
        ----------
        ax : matplotlib.Axes, optional
            Matplotliv axes object. Need to be a 3-array if flag2d is True
        flag2d : bool, default: ``False``
            Flag to either plot a 1D or 2D probability distributions
        x_component : int, optional
            Physical property to plot on the X-axis, as ordered in the GMM.
        y_component : int, optional
            Physical property to plot on the Y-axis, as ordered in the GMM
        padding : float, default: 0.2
            How much relative padding around the petrophysical means for the 1D and 2D plots
        plotting_precision : int, default: 100
            Number of divisions for the 1D and 2D plots
        plot_membership : bool, default: ``False``
            Plot the membership rather than the probability
        contour_opts : dict
            Modify the plotting options of the contour plot (in 1D and 2D)
        level_opts : dict
            Modify the plotting options of the level plot (in 1D and 2D)

        Returns
        -------
        matplotlib.Axes
            Axes including the plot
        """

        plotting_precision = int(plotting_precision)

        if x_component is None:
            x_component = 0
            if y_component is None:
                if flag2d and self.means_.shape[1] > 1:
                    y_component = 1

        if (not (x_component is None)) and (not (y_component is None)):
            flag2d = True

        if ax is None:
            if flag2d:
                fig = plt.figure(figsize=(10, 10))
                ax0 = plt.subplot2grid((4, 4), (3, 1), colspan=3)
                ax1 = plt.subplot2grid((4, 4), (0, 1), colspan=3, rowspan=3)
                ax2 = plt.subplot2grid((4, 4), (0, 0), rowspan=3)
                ax = [ax0, ax1, ax2]
            else:
                fig, ax = plt.subplots(1, 1, figsize=(8, 8))
            ax = np.r_[ax]

        # deal with the various possible shapes of covariances
        if self.covariance_type == "tied":
            covariances = np.r_[
                [self.covariances_ for i in range(self.n_components)]
            ].reshape(self.n_components, self.n_features_in_, self.n_features_in_)
        elif self.covariance_type == "diag" or self.covariance_type == "spherical":
            covariances = np.r_[
                [
                    self.covariances_[i] * np.eye(self.n_features_in_)
                    for i in range(self.n_components)
                ]
            ].reshape(self.n_components, self.n_features_in_, self.n_features_in_)
        else:
            covariances = self.covariances_

        dx = padding * (
            self.means_[:, x_component].max() - self.means_[:, x_component].min()
        )
        xmin, xmax = (
            self.means_[:, x_component].min() - dx,
            self.means_[:, x_component].max() + dx,
        )

        # create a sklearn.clustering.GaussianMixture for plotting (no influence from mesh and local weights)
        meansx = self.means_[:, x_component].reshape(self.n_components, 1)
        covx = covariances[:, [x_component]][:, :, [x_component]]
        if len(self.weights_.shape) == 2:
            weights = self.weights_.sum(axis=0)
            weights /= weights.sum()
        else:
            weights = self.weights_

        clfx = GaussianMixture(
            n_components=self.n_components,
            means_init=meansx,
            # limit computation to minimum as we set the model parameter a posteriori
            n_init=1,
            max_iter=2,
            # put a high tolerance to avoid warning about low model fit
            tol=1e256,
        )
        # random fit, we set values after.
        clfx.fit(np.random.randn(10, 1))
        clfx.means_ = meansx
        clfx.covariances_ = covx
        clfx.precisions_cholesky_ = _compute_precision_cholesky(
            clfx.covariances_, clfx.covariance_type
        )
        clfx.weights_ = weights

        xplot = np.linspace(xmin, xmax, plotting_precision)[:, np.newaxis]
        if plot_membership:
            rvx = clfx.predict(xplot)
            labelx = "membership"
        else:
            rvx = np.exp(clfx.score_samples(xplot))
            labelx = "1D Probability\nDensity\nDistribution"

        ax[0].set_xlim(xmin, xmax)
        ax[0].plot(
            xplot,
            rvx,
            linewidth=3.0,
            label=labelx,
            c="k",
        )
        ax[0].legend()
        ax[0].set_xlabel("Physical property {}".format(x_component))
        ax[0].set_ylabel("Probability Density values")

        if flag2d:
            dy = padding * (
                self.means_[:, y_component].max() - self.means_[:, y_component].min()
            )
            ymin, ymax = (
                self.means_[:, y_component].min() - dy,
                self.means_[:, y_component].max() + dy,
            )

            # create a sklearn.clustering.GaussianMixture for plotting (no influence from mesh and local weights)
            meansy = self.means_[:, y_component].reshape(self.n_components, 1)
            covy = covariances[:, [y_component]][:, :, [y_component]]

            clfy = GaussianMixture(
                n_components=self.n_components,
                means_init=meansy,
                # limit computation to minimum as we set the model parameter a posteriori
                n_init=1,
                max_iter=2,
                # put a high tolerance to avoid warning about low model fit
                tol=1e256,
            )
            # random fit, we set values after.
            clfy.fit(np.random.randn(10, 1))
            clfy.means_ = meansy
            clfy.covariances_ = covy
            clfy.precisions_cholesky_ = _compute_precision_cholesky(
                clfy.covariances_, clfy.covariance_type
            )
            clfy.weights_ = weights

            # 1d y-plot
            yplot = np.linspace(ymin, ymax, plotting_precision)[:, np.newaxis]
            if plot_membership:
                rvy = clfy.predict(yplot)
                labely = "membership"
            else:
                rvy = np.exp(clfy.score_samples(yplot))
                labely = "1D Probability\nDensity\nDistribution"
            ax[2].plot(rvy, yplot, linewidth=3.0, c="k", label=labely)
            ax[2].set_ylabel("Physical property {}".format(y_component))
            ax[2].set_ylim(ymin, ymax)
            ax[2].legend()

            # 2d plot
            mean2d = self.means_[:, [x_component, y_component]]
            cov2d = covariances[:, [x_component, y_component]][
                :, :, [x_component, y_component]
            ]
            clf2d = GaussianMixture(
                n_components=self.n_components,
                means_init=mean2d,
                n_init=1,
                max_iter=2,
                tol=1e256,
            )
            # random fit, we set values after.
            clf2d.fit(np.random.randn(10, 2))
            clf2d.means_ = mean2d
            clf2d.covariances_ = cov2d
            clf2d.precisions_cholesky_ = _compute_precision_cholesky(
                clf2d.covariances_, clf2d.covariance_type
            )
            clf2d.weights_ = weights

            x, y = np.mgrid[
                xmin : xmax : (xmax - xmin) / plotting_precision,
                ymin : ymax : (ymax - ymin) / plotting_precision,
            ]
            pos = np.empty(x.shape + (2,))
            pos[:, :, 0] = x
            pos[:, :, 1] = y

            if plot_membership:
                rv2d = clf2d.predict(pos.reshape(-1, 2))
                labely = "membership"
            else:
                rv2d = clf2d.score_samples(pos.reshape(-1, 2))
                labely = "2D Probability Density Distribution"

            if contour_opts is None:
                contour_opts = {}
            contour_opts = {"levels": 10, "cmap": "viridis", **contour_opts}
            surf = ax[1].contourf(x, y, rv2d.reshape(x.shape), **contour_opts)

            if level_opts is None:
                level_opts = {}
            level_opts = {
                "levels": 10,
                "colors": "k",
                "linewidths": 1.0,
                "linestyles": "dashdot",
                **level_opts,
            }

            ax[1].contour(x, y, rv2d.reshape(x.shape), **level_opts)
            ax[1].scatter(
                meansx,
                meansy,
                label="Petrophysical means",
                cmap="inferno_r",
                c=np.linspace(0, self.n_components, self.n_components),
                marker="v",
                edgecolors="k",
            )

            axbar = inset_axes(
                ax[1],
                width="40%",
                height="3%",
                loc="upper right",
                borderpad=1,
            )
            cbpetro = plt.colorbar(surf, cax=axbar, orientation="horizontal")
            cbpetro.set_ticks([rv2d.min(), rv2d.max()])
            cbpetro.set_ticklabels(["Low", "High"])
            cbpetro.set_label(labely)
            cbpetro.outline.set_edgecolor("k")

            ax[1].set_xlim(xmin, xmax)
            ax[1].set_ylim(ymin, ymax)
            ax[1].legend(loc=3)
            ax[1].set_ylabel("")
            ax[1].set_xlabel("")

        return ax


class GaussianMixtureWithPrior(WeightedGaussianMixture):
    """
    This class built upon the WeightedGaussianMixture, which itself built upon from
    the mixture.gaussian_mixture.GaussianMixture class from Scikit-Learn.

    In addition to weights samples/observations by the cells volume of the mesh,
    this class uses a posterior approach to fit the GMM parameters. This means
    it takes prior parameters, passed through `WeightedGaussianMixture` gmmref.
    The prior distribution for each parameters (proportions, means, covariances) is
    defined through a conjugate or semi-conjugate approach (prior_type), to the choice of the user.
    See Astic & Oldenburg 2019: A framework for petrophysically and geologically
    guided geophysical inversion (https://doi.org/10.1093/gji/ggz389) for more information.

    Disclaimer: this class built upon the GaussianMixture class from Scikit-Learn.
    New functionalitie are added, as well as modifications to
    existing functions, to serve the purposes pursued within SimPEG.
    This use is allowed by the Scikit-Learn licensing (BSD-3-Clause License)
    and we are grateful for their contributions to the open-source community.

    Addtional parameters to provide, compared to `WeightedGaussianMixture`:

    Parameters
    ----------
    kappa : numpy.ndarray
        strength of the confidence in the prior means
    nu : numpy.ndarry
        strength of the confidence in the prior covariances
    zeta : numpy.ndarry
        strength of the confidence in the prior proportions
    prior_type : str
        Choose from one of the following:

            - "semi": semi-conjugate prior, the means and covariances priors are indepedent
            - "full": conjugate prior, the means and covariances priors are inter-depedent

    update_covariances : bool
        Choose from two options:

        - ``True``: semi or conjugate prior by averaging the covariances
        - ``False``: alternative (not conjugate) prior: average the precisions instead

    fixed_membership : numpy.ndarray of int, optional
        A 2d numpy.ndarray to fix the membership to a chosen lithology of particular cells.
        The first column contains the numeric index of the cells, the second column the respective lithology index.
        Shape is (index of the fixed cell, lithology index) fixed_membership:
    """

    def __init__(
        self,
        gmmref,
        kappa=0.0,
        nu=0.0,
        zeta=0.0,
        prior_type="semi",  # semi or full
        update_covariances=True,
        fixed_membership=None,
        init_params="kmeans",
        max_iter=100,
        means_init=None,
        n_init=10,
        precisions_init=None,
        random_state=None,
        reg_covar=1e-06,
        tol=0.001,
        verbose=0,
        verbose_interval=10,
        warm_start=False,
        weights_init=None,
        # **kwargs
    ):
        self.mesh = gmmref.mesh
        self.n_components = gmmref.n_components
        self.gmmref = gmmref
        self.covariance_type = gmmref.covariance_type
        self.kappa = kappa * np.ones((self.n_components, gmmref.means_.shape[1]))
        self.nu = nu * np.ones(self.n_components)
        self.zeta = zeta * np.ones_like(self.gmmref.weights_)
        self.prior_type = prior_type
        self.update_covariances = update_covariances
        self.fixed_membership = fixed_membership

        super(GaussianMixtureWithPrior, self).__init__(
            covariance_type=self.covariance_type,
            mesh=self.mesh,
            actv=self.gmmref.actv,
            init_params=init_params,
            max_iter=max_iter,
            means_init=means_init,
            n_components=self.n_components,
            n_init=n_init,
            precisions_init=precisions_init,
            random_state=random_state,
            reg_covar=reg_covar,
            tol=tol,
            verbose=verbose,
            verbose_interval=verbose_interval,
            warm_start=warm_start,
            weights_init=weights_init,
            # **kwargs
        )
        # set_kwargs(self, **kwargs)

    def order_cluster(self, outputindex=False):
        """Order cluster

        Arrange the clusters of gmm in the same order as those of gmmref,
        based on their relative similarities and priorizing first the most proeminent
        clusters (highest proportions)

        Parameters
        ----------
        outputindex : bool, default: ``False``
            If ``True``, return the ordering index for the clusters of GMM.

        Returns
        -------
        numpy.ndarray of int
            Returns the ordering index for the clusters of GMM if
            *outputindex* is ``True``
        """
        self.order_clusters_GM_weight()

        idx_ref = np.ones(len(self.gmmref.means_), dtype=bool)

        indx = []

        for i in range(self.n_components):
            dis = self._estimate_log_prob(
                self.gmmref.means_[idx_ref].reshape(
                    [-1] + [d for d in self.gmmref.means_.shape[1:]]
                )
            )
            id_dis = dis.argmax(axis=0)[i]
            idrefmean = np.where(
                np.all(
                    self.gmmref.means_ == self.gmmref.means_[idx_ref][id_dis], axis=1
                )
            )[0][0]
            indx.append(idrefmean)
            idx_ref[idrefmean] = False

        self.means_ = self.means_[indx].reshape(self.means_.shape)

        if self.weights_.ndim == 1:
            self.weights_ = self.weights_[indx].reshape(self.weights_.shape)
        else:
            self.weights_ = self.weights_[:, indx].reshape(self.weights_.shape)

        if self.covariance_type == "tied":
            pass
        else:
            self.precisions_ = self.precisions_[indx].reshape(self.precisions_.shape)
            self.covariances_ = self.covariances_[indx].reshape(self.covariances_.shape)

        self.precisions_cholesky_ = _compute_precision_cholesky(
            self.covariances_, self.covariance_type
        )

        if outputindex:
            return indx

    def update_gmm_with_priors(self, debug=False):
        """Update GMM with priors

        This method, inserted in the fit function, modify the Maximum Likelyhood Estimation (MLE)
        of the GMM's parameters to a Maximum A Posteriori (MAP) estimation.

        Parameters
        ----------
        debug : bool, default: ``False``
            Print debug statments
        """

        self.compute_clusters_precisions()
        self.order_cluster()

        if debug:
            print("before update means: ", self.means_)
            print("before update weights: ", self.weights_)
            print("before update precisions: ", self.precisions_)

        if self.weights_.ndim == 1:
            weights_ = self.weights_
        else:
            weights_ = (np.c_[self.cell_volumes] * self.weights_).sum(axis=0) / (
                np.c_[self.cell_volumes] * self.weights_
            ).sum()

        if self.gmmref.weights_.ndim == 1:
            ref_weights_ = self.gmmref.weights_
        else:
            ref_weights_ = (np.c_[self.gmmref.cell_volumes] * self.gmmref.weights_).sum(
                axis=0
            ) / (np.c_[self.gmmref.cell_volumes] * self.gmmref.weights_).sum()

        for k in range(self.n_components):
            if self.prior_type == "full":
                smu = (self.kappa[k] * weights_[k]) * (
                    (self.gmmref.means_[k] - self.means_[k]) ** 2.0
                )
                smu /= self.kappa[k] + weights_[k]
                smu *= 1.0 / (weights_[k] + ref_weights_[k] * self.nu[k])

            self.means_[k] = (1.0 / (weights_[k] + ref_weights_[k] * self.kappa[k])) * (
                weights_[k] * self.means_[k]
                + ref_weights_[k] * self.kappa[k] * self.gmmref.means_[k]
            )

            if self.gmmref.covariance_type == "tied":
                pass
            elif self.update_covariances:
                self.covariances_[k] = (
                    1.0 / (weights_[k] + ref_weights_[k] * self.nu[k])
                ) * (
                    weights_[k] * self.covariances_[k]
                    + ref_weights_[k] * self.nu[k] * self.gmmref.covariances_[k]
                )

                if self.prior_type == "full":
                    self.covariances_[k] += smu

            else:
                self.precisions_[k] = (
                    1.0 / (weights_[k] + ref_weights_[k] * self.nu[k])
                ) * (
                    weights_[k] * self.precisions_[k]
                    + ref_weights_[k] * self.nu[k] * self.gmmref.precisions_[k]
                )

        self.weights_ = (
            (self.weights_ + self.zeta * self.gmmref.weights_).T
            * (1.0 / (1.0 + np.sum((self.zeta * self.gmmref.weights_).T, axis=0)))
        ).T

        if self.gmmref.covariance_type == "tied":
            if self.update_covariances:
                self.covariances_ = (1.0 / (1.0 + np.sum(ref_weights_ * self.nu))) * (
                    self.covariances_
                    + np.sum(ref_weights_ * self.nu) * self.gmmref.covariances_
                )
                self.compute_clusters_precisions()
            else:
                self.precisions_ = (1.0 / (1.0 + np.sum(ref_weights_ * self.nu))) * (
                    self.precisions_
                    + np.sum(ref_weights_ * self.nu) * self.gmmref.precisions_
                )
                self.compute_clusters_covariances()
        elif self.update_covariances:
            self.compute_clusters_precisions()
        else:
            self.compute_clusters_covariances()

        if debug:
            print("after update means: ", self.means_)
            print("after update weights: ", self.weights_)
            print("after update precisions: ", self.precisions_)

    def fit(self, X, y=None, debug=False):
        """Estimate model parameters with the EM algorithm.

        [modified from Scikit-Learn for Maximum A Posteriori estimates (MAP)]
        The method fits the model ``n_init`` times and sets the parameters with
        which the model has the largest likelihood or lower bound. Within each
        trial, the method iterates between E-step and M-step for ``max_iter``
        times until the change of likelihood or lower bound is less than
        ``tol``, otherwise, a ``ConvergenceWarning`` is raised.
        If ``warm_start`` is ``True``, then ``n_init`` is ignored and a single
        initialization is performed upon the first call. Upon consecutive
        calls, training starts where it left off.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            List of n_features-dimensional data points. Each row
            corresponds to a single data point.
        y : Ignored
            Not used, present for API consistency by convention.
        debug : bool, default: ``False``
            If ``True``, print debug statements

        Returns
        -------
        self : object
            The fitted mixture.
        """
        self.fit_predict(X, y, debug)
        return self

    def fit_predict(self, X, y=None, debug=False):
        """Estimate model parameters using X and predict the labels for X.

        [modified from Scikit-Learn for Maximum A Posteriori estimates (MAP)]
        The method fits the model n_init times and sets the parameters with
        which the model has the largest likelihood or lower bound. Within each
        trial, the method iterates between E-step and M-step for `max_iter`
        times until the change of likelihood or lower bound is less than
        `tol`, otherwise, a :class:`~sklearn.exceptions.ConvergenceWarning` is
        raised. After fitting, it predicts the most probable label for the
        input data points.

        Parameters
        ----------
        X : (n_samples, n_features) array_like
            List of n_features-dimensional data points. Each row
            corresponds to a single data point.
        y : Ignored
            Not used, present for API consistency by convention.
        debug : bool, default: ``False``
            If ``True``, print debug statements

        Returns
        -------
        (n_samples) array
            Component labels.
        """
        if self.verbose:
            print("modified from scikit-learn")

        X = self._validate_data(X, dtype=[np.float64, np.float32], ensure_min_samples=2)
        if X.shape[0] < self.n_components:
            raise ValueError(
                "Expected n_samples >= n_components "
                f"but got n_components = {self.n_components}, "
                f"n_samples = {X.shape[0]}"
            )
        self._check_parameters(X)

        # if we enable warm_start, we will have a unique initialisation
        do_init = not (self.warm_start and hasattr(self, "converged_"))
        n_init = self.n_init if do_init else 1

        max_lower_bound = -np.inf
        self.converged_ = False

        random_state = check_random_state(self.random_state)

        n_samples, _ = X.shape
        for init in range(n_init):
            self._print_verbose_msg_init_beg(init)

            if do_init:
                self._initialize_parameters(X, random_state)

            lower_bound = -np.inf if do_init else self.lower_bound_

            for n_iter in range(1, self.max_iter + 1):
                prev_lower_bound = lower_bound

                log_prob_norm, log_resp = self._e_step(X)

                if self.fixed_membership is not None:
                    # force responsibilities
                    aux = -(np.inf) * np.ones(
                        (self.fixed_membership.shape[0], self.n_components)
                    )
                    aux[np.arange(len(aux)), self.fixed_membership[:, 1]] = 0.0
                    log_resp[self.fixed_membership[:, 0]] = aux

                self._m_step(X, log_resp)
                self.update_gmm_with_priors(debug=debug)

                if self.fixed_membership is not None and self.weights_.ndim == 2:
                    # force local weights
                    aux = np.zeros((self.fixed_membership.shape[0], self.n_components))
                    aux[np.arange(len(aux)), self.fixed_membership[:, 1]] = 1
                    self.weights_[self.fixed_membership[:, 0]] = aux

                lower_bound = self._compute_lower_bound(log_resp, log_prob_norm)

                change = lower_bound - prev_lower_bound
                self._print_verbose_msg_iter_end(n_iter, change)

                if abs(change) < self.tol:
                    self.converged_ = True
                    break

            self._print_verbose_msg_init_end(lower_bound)

            if lower_bound > max_lower_bound or max_lower_bound == -np.inf:
                max_lower_bound = lower_bound
                best_params = self._get_parameters()
                best_n_iter = n_iter

        if not self.converged_:
            warnings.warn(
                "Initialization %d did not converge. "
                "Try different init parameters, "
                "or increase max_iter, tol "
                "or check for degenerate data." % (init + 1),
                ConvergenceWarning,
            )

        self._set_parameters(best_params)
        self.n_iter_ = best_n_iter
        self.lower_bound_ = max_lower_bound

        return self


class GaussianMixtureWithNonlinearRelationships(WeightedGaussianMixture):
    """Gaussian mixture class for non-linear relationships.

    This class built upon the WeightedGaussianMixture, which itself built upon from
    the mixture.gaussian_mixture.GaussianMixture class from Scikit-Learn.

    In addition to weights samples/observations by the cells volume of the mesh,
    this class can be given specified nonlinear relationships between physical properties.
    (polynomial etc.) Those nonlinear relationships are given in the form of a
    list of mapping (cluster_mapping argument). Functions are added and modified
    to fill that purpose, in particular the `fit` and  `samples` functions.

    Disclaimer: this class built upon the GaussianMixture class from Scikit-Learn.
    New functionalitie are added, as well as modifications to
    existing functions, to serve the purposes pursued within SimPEG.
    This use is allowed by the Scikit-Learn licensing (BSD-3-Clause License)
    and we are grateful for their contributions to the open-source community.

    Addtional parameters to provide, compared to `WeightedGaussianMixture`:

    Parameters
    ----------
    cluster_mapping : (n_components) list
        List of mapping describing a nonlinear relationships between physical properties; one per cluster/unit.
    """

    def __init__(
        self,
        mesh,
        n_components=1,
        covariance_type="full",
        tol=1e-3,
        reg_covar=1e-6,
        max_iter=100,
        n_init=1,
        init_params="kmeans",
        weights_init=None,
        means_init=None,
        precisions_init=None,
        random_state=None,
        warm_start=False,
        verbose=0,
        verbose_interval=10,
        cluster_mapping=None,
    ):
        if cluster_mapping is None:
            self.cluster_mapping = [IdentityMap() for i in range(n_components)]
        else:
            self.cluster_mapping = cluster_mapping

        super(GaussianMixtureWithNonlinearRelationships, self).__init__(
            mesh=mesh,
            covariance_type=covariance_type,
            init_params=init_params,
            max_iter=max_iter,
            means_init=means_init,
            n_components=n_components,
            n_init=n_init,
            precisions_init=precisions_init,
            random_state=random_state,
            reg_covar=reg_covar,
            tol=tol,
            verbose=verbose,
            verbose_interval=verbose_interval,
            warm_start=warm_start,
            weights_init=weights_init,
            # **kwargs
        )

    def _initialize(self, X, resp):
        """
        [modified from Scikit-Learn.mixture.gaussian_mixture]
        Initialization of the Gaussian mixture parameters.
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
        resp : array-like, shape (n_samples, n_components)
        """
        n_samples, _ = X.shape

        weights, means, covariances = self._estimate_gaussian_parameters(
            X, resp, self.reg_covar, self.covariance_type
        )
        weights /= n_samples

        self.weights_ = weights if self.weights_init is None else self.weights_init
        self.means_ = means if self.means_init is None else self.means_init

        if self.precisions_init is None:
            self.covariances_ = covariances
            self.precisions_cholesky_ = _compute_precision_cholesky(
                covariances, self.covariance_type
            )
        elif self.covariance_type == "full":
            self.precisions_cholesky_ = np.array(
                [
                    linalg.cholesky(prec_init, lower=True)
                    for prec_init in self.precisions_init
                ]
            )
        elif self.covariance_type == "tied":
            self.precisions_cholesky_ = linalg.cholesky(
                self.precisions_init, lower=True
            )
        else:
            self.precisions_cholesky_ = self.precisions_init

    @classmethod
    def _estimate_log_gaussian_prob(
        self, X, means, precisions_chol, covariance_type, cluster_mapping
    ):
        """
        [modified from Scikit-Learn.mixture.gaussian_mixture]
        Estimate the log Gaussian probability.
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
        means : array-like, shape (n_components, n_features)
        precisions_chol : array-like
            Cholesky decompositions of the precision matrices.
            'full' : shape of (n_components, n_features, n_features)
            'tied' : shape of (n_features, n_features)
            'diag' : shape of (n_components, n_features)
            'spherical' : shape of (n_components,)
        covariance_type : {'full', 'tied', 'diag', 'spherical'}
        Returns
        -------
        log_prob : array, shape (n_samples, n_components)
        """
        n_samples, n_features = X.shape
        n_components, _ = means.shape
        # det(precision_chol) is half of det(precision)
        log_det = _compute_log_det_cholesky(
            precisions_chol, covariance_type, n_features
        )

        if covariance_type == "full":
            log_prob = np.empty((n_samples, n_components))
            for k, (mu, prec_chol, mapping) in enumerate(
                zip(means, precisions_chol, cluster_mapping)
            ):
                y = np.dot(mapping * X, prec_chol) - np.dot(mu, prec_chol)
                log_prob[:, k] = np.sum(np.square(y), axis=1)

        elif covariance_type == "tied":
            log_prob = np.empty((n_samples, n_components))
            for k, (mu, mapping) in enumerate(zip(means, cluster_mapping)):
                y = np.dot(mapping * X, precisions_chol) - np.dot(mu, precisions_chol)
                log_prob[:, k] = np.sum(np.square(y), axis=1)

        elif covariance_type == "diag" or covariance_type == "spherical":
            log_prob = np.empty((n_samples, n_components))
            for k, (mu, prec_chol, mapping) in enumerate(
                zip(means, precisions_chol, cluster_mapping)
            ):
                y = np.dot(mapping * X, prec_chol * np.eye(n_features)) - np.dot(
                    mu, prec_chol * np.eye(n_features)
                )
                log_prob[:, k] = np.sum(np.square(y), axis=1)

        return -0.5 * (n_features * np.log(2 * np.pi) + log_prob) + log_det

    def _estimate_log_prob(self, X):
        """
        [modified from Scikit-Learn.mixture.gaussian_mixture]
        """
        return self._estimate_log_gaussian_prob(
            X,
            self.means_,
            self.precisions_cholesky_,
            self.covariance_type,
            self.cluster_mapping,
        )

    def _estimate_gaussian_parameters(self, X, resp, reg_covar, covariance_type):
        """
        [modified from Scikit-Learn.mixture.gaussian_mixture]
        """
        respVol = self.cell_volumes.reshape(-1, 1) * resp
        nk = respVol.sum(axis=0) + 10 * np.finfo(resp.dtype).eps
        # stupid lazy piece of junk code to get the shapes right
        means = np.dot(respVol.T, X) / nk[:, np.newaxis]
        covariances = {
            "full": _estimate_gaussian_covariances_full,
            "tied": self._estimate_gaussian_covariances_tied,
            "diag": _estimate_gaussian_covariances_diag,
            "spherical": _estimate_gaussian_covariances_spherical,
        }[covariance_type](respVol, X, nk, means, reg_covar)
        # The actual calculation
        for k in range(means.shape[0]):
            means[k] = (
                np.dot(respVol.T, self.cluster_mapping[k] * X) / nk[:, np.newaxis]
            )[k]
        for k in range(means.shape[0]):
            covariances[k] = (
                {
                    "full": _estimate_gaussian_covariances_full,
                    "tied": self._estimate_gaussian_covariances_tied,
                    "diag": _estimate_gaussian_covariances_diag,
                    "spherical": _estimate_gaussian_covariances_spherical,
                }[covariance_type](
                    respVol, self.cluster_mapping[k] * X, nk, means, reg_covar
                )
            )[k]
        return nk, means, covariances

    # TODOs: Still not working because of inverse mapping not implemented
    def sample(self, n_samples=1):
        """
        [modified from Scikit-Learn.mixture.gaussian_mixture]
        Generate random samples from the fitted Gaussian distribution
        with nonlinear relationships.

        Parameters
        ----------
        n_samples : int, default: `
            Number of samples to generate.

        Returns
        -------
        X : array, shape (n_samples, n_features)
            Randomly generated sample
        y : array, shape (nsamples,)
            Component labels

        """
        check_is_fitted(self)

        if n_samples < 1:
            raise ValueError(
                "Invalid value for 'n_samples': %d . The sampling requires at "
                "least one sample." % (self.n_components)
            )

        _, n_features = self.means_.shape
        rng = check_random_state(self.random_state)
        n_samples_comp = rng.multinomial(n_samples, self.weights_)

        if self.covariance_type == "full":
            X = np.vstack(
                [
                    rng.multivariate_normal(mean, covariance, int(sample))
                    for (mean, covariance, sample) in zip(
                        self.means_, self.covariances_, n_samples_comp
                    )
                ]
            )
        elif self.covariance_type == "tied":
            X = np.vstack(
                [
                    rng.multivariate_normal(mean, self.covariances_, int(sample))
                    for (mean, sample) in zip(self.means_, n_samples_comp)
                ]
            )
        else:
            X = np.vstack(
                [
                    mean + rng.randn(sample, n_features) * np.sqrt(covariance)
                    for (mean, covariance, sample) in zip(
                        self.means_, self.covariances_, n_samples_comp
                    )
                ]
            )

        y = np.concatenate(
            [j * np.ones(sample, dtype=int) for j, sample in enumerate(n_samples_comp)]
        )
        X = np.vstack(
            [
                self.cluster_mapping[y[i]].inverse(X[i].reshape(-1, n_features))
                for i in range(len(X))
            ]
        )

        return (X, y)

    def _m_step(self, X, log_resp):
        """
        [modified from Scikit-Learn.mixture.gaussian_mixture]
        M step.
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
        log_resp : array-like, shape (n_samples, n_components)
            Logarithm of the posterior probabilities (or responsibilities) of
            the point of each sample in X.
        """
        n_samples, _ = X.shape
        (
            self.weights_,
            self.means_,
            self.covariances_,
        ) = self._estimate_gaussian_parameters(
            X, np.exp(log_resp), self.reg_covar, self.covariance_type
        )
        self.weights_ /= n_samples
        self.precisions_cholesky_ = _compute_precision_cholesky(
            self.covariances_, self.covariance_type
        )


class GaussianMixtureWithNonlinearRelationshipsWithPrior(GaussianMixtureWithPrior):
    """Gaussian mixture class for non-linear relationships with priors.

    This class built upon the `GaussianMixtureWithPrior`, which itself built upon from
    the `WeightedGaussianMixture`, built up from the
    mixture.gaussian_mixture.GaussianMixture class from Scikit-Learn.

    In addition to weights samples/observations by the cells volume of the mesh
    (from `WeightedGaussianMixture`), and nonlinear relationships for each cluster
    (from `GaussianMixtureWithNonlinearRelationships`), this class uses a
    posterior approach to fit the GMM parameters (from `GaussianMixtureWithPrior`).
    It takes prior parameters, passed through `WeightedGaussianMixture` gmmref.
    The prior distribution for each parameters (proportions, means, covariances) is
    defined through a conjugate or semi-conjugate approach (prior_type), to the choice of the user.
    See Astic & Oldenburg 2019: A framework for petrophysically and geologically
    guided geophysical inversion (https://doi.org/10.1093/gji/ggz389) for more information.

    Disclaimer: this class built upon the GaussianMixture class from Scikit-Learn.
    New functionalitie are added, as well as modifications to
    existing functions, to serve the purposes pursued within SimPEG.
    This use is allowed by the Scikit-Learn licensing (BSD-3-Clause License)
    and we are grateful for their contributions to the open-source community.

    Addtional parameters to provide, compared to `GaussianMixtureWithPrior`:

    Parameters
    ----------
    cluster_mapping : (n_components) list
        List of mapping describing a nonlinear relationships between physical properties; one per cluster/unit.

    """

    def __init__(
        self,
        gmmref,
        kappa=0.0,
        nu=0.0,
        zeta=0.0,
        prior_type="semi",  # semi or conjugate
        cluster_mapping=None,
        init_params="kmeans",
        max_iter=100,
        means_init=None,
        n_init=10,
        precisions_init=None,
        random_state=None,
        reg_covar=1e-06,
        tol=0.001,
        verbose=0,
        verbose_interval=10,
        warm_start=False,
        weights_init=None,
        update_covariances=True,
        fixed_membership=None,
    ):
        if cluster_mapping is None:
            self.cluster_mapping = gmmref.cluster_mapping
        else:
            self.cluster_mapping = cluster_mapping

        super(GaussianMixtureWithNonlinearRelationshipsWithPrior, self).__init__(
            gmmref=gmmref,
            kappa=kappa,
            nu=nu,
            zeta=zeta,
            init_params=init_params,
            max_iter=max_iter,
            means_init=means_init,
            n_init=n_init,
            precisions_init=precisions_init,
            random_state=random_state,
            reg_covar=reg_covar,
            tol=tol,
            verbose=verbose,
            verbose_interval=verbose_interval,
            warm_start=warm_start,
            weights_init=weights_init,
            update_covariances=update_covariances,
            fixed_membership=fixed_membership
            # **kwargs
        )

    def _initialize(self, X, resp):
        """
        [modified from Scikit-Learn.mixture.gaussian_mixture]
        Initialization of the Gaussian mixture parameters.
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
        resp : array-like, shape (n_samples, n_components)
        """
        n_samples, _ = X.shape

        weights, means, covariances = self._estimate_gaussian_parameters(
            X, resp, self.reg_covar, self.covariance_type
        )
        weights /= n_samples

        self.weights_ = weights if self.weights_init is None else self.weights_init
        self.means_ = means if self.means_init is None else self.means_init

        if self.precisions_init is None:
            self.covariances_ = covariances
            self.precisions_cholesky_ = _compute_precision_cholesky(
                covariances, self.covariance_type
            )
        elif self.covariance_type == "full":
            self.precisions_cholesky_ = np.array(
                [
                    linalg.cholesky(prec_init, lower=True)
                    for prec_init in self.precisions_init
                ]
            )
        elif self.covariance_type == "tied":
            self.precisions_cholesky_ = linalg.cholesky(
                self.precisions_init, lower=True
            )
        else:
            self.precisions_cholesky_ = self.precisions_init

    @classmethod
    def _estimate_log_gaussian_prob(
        self, X, means, precisions_chol, covariance_type, cluster_mapping
    ):
        """
        [modified from Scikit-Learn.mixture.gaussian_mixture]
        Estimate the log Gaussian probability.
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
        means : array-like, shape (n_components, n_features)
        precisions_chol : array-like
            Cholesky decompositions of the precision matrices.
            'full' : shape of (n_components, n_features, n_features)
            'tied' : shape of (n_features, n_features)
            'diag' : shape of (n_components, n_features)
            'spherical' : shape of (n_components,)
        covariance_type : {'full', 'tied', 'diag', 'spherical'}
        cluster_mapping: list of mapping of length (n_components,)
        Returns
        -------
        log_prob : array, shape (n_samples, n_components)
        """
        n_samples, n_features = X.shape
        n_components, _ = means.shape
        # det(precision_chol) is half of det(precision)
        log_det = _compute_log_det_cholesky(
            precisions_chol, covariance_type, n_features
        )

        if covariance_type == "full":
            log_prob = np.empty((n_samples, n_components))
            for k, (mu, prec_chol, mapping) in enumerate(
                zip(means, precisions_chol, cluster_mapping)
            ):
                y = np.dot(mapping * X, prec_chol) - np.dot(mu, prec_chol)
                log_prob[:, k] = np.sum(np.square(y), axis=1)

        elif covariance_type == "tied":
            log_prob = np.empty((n_samples, n_components))
            for k, (mu, mapping) in enumerate(zip(means, cluster_mapping)):
                y = np.dot(mapping * X, precisions_chol) - np.dot(mu, precisions_chol)
                log_prob[:, k] = np.sum(np.square(y), axis=1)

        elif covariance_type == "diag" or covariance_type == "spherical":
            log_prob = np.empty((n_samples, n_components))
            for k, (mu, prec_chol, mapping) in enumerate(
                zip(means, precisions_chol, cluster_mapping)
            ):
                y = np.dot(mapping * X, prec_chol * np.eye(n_features)) - np.dot(
                    mu, prec_chol * np.eye(n_features)
                )
                log_prob[:, k] = np.sum(np.square(y), axis=1)

        return -0.5 * (n_features * np.log(2 * np.pi) + log_prob) + log_det

    def _estimate_log_prob(self, X):
        return self._estimate_log_gaussian_prob(
            X,
            self.means_,
            self.precisions_cholesky_,
            self.covariance_type,
            self.cluster_mapping,
        )

    def _estimate_gaussian_parameters(self, X, resp, reg_covar, covariance_type):
        """
        [modified from Scikit-Learn.mixture.gaussian_mixture]
        """
        respVol = self.cell_volumes.reshape(-1, 1) * resp
        nk = respVol.sum(axis=0) + 10 * np.finfo(resp.dtype).eps
        # stupid lazy piece of junk code to get the shapes right
        means = np.dot(respVol.T, X) / nk[:, np.newaxis]
        covariances = {
            "full": _estimate_gaussian_covariances_full,
            "tied": self._estimate_gaussian_covariances_tied,
            "diag": _estimate_gaussian_covariances_diag,
            "spherical": _estimate_gaussian_covariances_spherical,
        }[covariance_type](respVol, X, nk, means, reg_covar)
        # The actual calculation
        for k in range(means.shape[0]):
            means[k] = (
                np.dot(respVol.T, self.cluster_mapping[k] * X) / nk[:, np.newaxis]
            )[k]
        for k in range(means.shape[0]):
            covariances[k] = (
                {
                    "full": _estimate_gaussian_covariances_full,
                    "tied": self._estimate_gaussian_covariances_tied,
                    "diag": _estimate_gaussian_covariances_diag,
                    "spherical": _estimate_gaussian_covariances_spherical,
                }[covariance_type](
                    respVol, self.cluster_mapping[k] * X, nk, means, reg_covar
                )
            )[k]
        return nk, means, covariances

    def _m_step(self, X, log_resp):
        """
        [modified from Scikit-Learn.mixture.gaussian_mixture]
        M step.
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
        log_resp : array-like, shape (n_samples, n_components)
            Logarithm of the posterior probabilities (or responsibilities) of
            the point of each sample in X.
        """
        n_samples, _ = X.shape
        (
            self.weights_,
            self.means_,
            self.covariances_,
        ) = self._estimate_gaussian_parameters(
            X, np.exp(log_resp), self.reg_covar, self.covariance_type
        )
        self.weights_ /= n_samples
        self.precisions_cholesky_ = _compute_precision_cholesky(
            self.covariances_, self.covariance_type
        )



from ..objective_function import ComboObjectiveFunction, BaseObjectiveFunction
import dask
import dask.array as da
import os
import shutil
import numpy as np
from dask.distributed import Future, get_client, Client
from ..data_misfit import L2DataMisfit
BaseObjectiveFunction._workers = None


@property
def client(self):
    if getattr(self, '_client', None) is None:
        self._client = get_client()

    return self._client


@client.setter
def client(self, client):
    assert isinstance(client, Client)
    self._client = client


BaseObjectiveFunction.client = client


@property
def workers(self):
    return self._workers


@workers.setter
def workers(self, workers):
    self._workers = workers


BaseObjectiveFunction.workers = workers


def dask_call(self, m, f=None):
    fcts = []
    multipliers = []
    for i, phi in enumerate(self):
        multiplier, objfct = phi
        if multiplier == 0.0:  # don't evaluate the fct
            continue
        else:

            if f is not None and objfct._has_fields:
                fct = objfct(m, f=f[i])
            else:
                fct = objfct(m)

            if isinstance(fct, Future):
                future = self.client.compute(
                    self.client.submit(da.multiply, multiplier, fct).result()
                )
                fcts += [future]
            else:
                fcts += [fct]

            multipliers += [multiplier]

    if isinstance(fcts[0], Future):
        phi = self.client.submit(
            da.sum, self.client.submit(da.vstack, fcts), axis=0
        ).result()
        return phi
    else:
        return np.sum(
            np.r_[multipliers][:, None] * np.vstack(fcts), axis=0
        ).squeeze()


ComboObjectiveFunction.__call__ = dask_call


def dask_deriv(self, m, f=None):
    """
    First derivative of the composite objective function is the sum of the
    derivatives of each objective function in the list, weighted by their
    respective multplier.

    :param numpy.ndarray m: model
    :param SimPEG.Fields f: Fields object (if applicable)
    """

    g = []
    multipliers = []
    for i, phi in enumerate(self):
        multiplier, objfct = phi
        if multiplier == 0.0:  # don't evaluate the fct
            continue
        else:

            if f is not None and isinstance(objfct, L2DataMisfit):
                fct = objfct.deriv(m, f=f[i])
            else:
                fct = objfct.deriv(m)

            if isinstance(fct, Future):
                future = self.client.compute(
                    self.client.submit(da.multiply, multiplier, fct)
                )
                g += [future]
            else:
                g += [fct]

            multipliers += [multiplier]

    if isinstance(g[0], Future):
        big_future = self.client.submit(
            da.sum, self.client.submit(da.vstack, g), axis=0
        ).result()
        return self.client.compute(big_future).result()

    else:
        return np.sum(
            np.r_[multipliers][:, None] * np.vstack(g), axis=0
        ).squeeze()


ComboObjectiveFunction.deriv = dask_deriv


def dask_deriv2(self, m, v=None, f=None):
    """
    Second derivative of the composite objective function is the sum of the
    second derivatives of each objective function in the list, weighted by
    their respective multplier.

    :param numpy.ndarray m: model
    :param numpy.ndarray v: vector we are multiplying by
    :param SimPEG.Fields f: Fields object (if applicable)
    """

    H = []
    multipliers = []
    for i, phi in enumerate(self):
        multiplier, objfct = phi
        if multiplier == 0.0:  # don't evaluate the fct
            continue
        else:
            fct = objfct.deriv2(m, v)

            if isinstance(fct, Future):
                future = self.client.submit(da.multiply, multiplier, fct)
                H += [future]
            else:
                H += [fct]

            multipliers += [multiplier]

    if isinstance(H[0], Future):
        big_future = self.client.submit(
            da.sum, self.client.submit(da.vstack, H), axis=0
        ).result()

        return np.asarray(big_future)

    else:
        phi_deriv2 = 0
        for multiplier, h in zip(multipliers, H):
            phi_deriv2 += multiplier * h

        return phi_deriv2


ComboObjectiveFunction.deriv2 = dask_deriv2



# Interactive NSEM map plot
import numpy as np
import matplotlib.pyplot as plt

from .. import receivers as rx

from .plot_utils import ApparentResPhsStationPlot, TipperAmplitudeStationPlot


class NSEM_data_viewer:
    """
    An interactive SimPEG NSEM Data viewer.

    Generates a clickble location map of the data, plotting data curves
    in a separate window.

    :param SimPEG.electromagnetics.natural_source.Data data: Data object, needs to have assigned
        relative_error and noise_floor

    :param data_dict: A dictionary of other NSEM Data objects
    :type data_dict: dict, optional
    :param backend: Flag to control the backend used in matplotlib
    :type backend: str, optional

    """

    _need_setup = True

    def __init__(self, data, data_dict=None, backend="qt"):
        """
        Initialize an object

        """
        # Set data
        self._data = data
        self._data_dict = data_dict

        # Set the default component
        if data_dict is not None:
            self.dict_comp = data_dict.keys()[0]
        else:
            self._dict_comp = None

        # Open the location figure
        self.location_fig, self._location_ax = self._data.map_data_locations(picker=5)
        # Seth the onClick event function
        self.location_fig.canvas.mpl_connect("pick_event", self._onClick)

        self.text = self._location_ax.text(
            0.05,
            1.05,
            "Selected: none",
            transform=self._location_ax.transAxes,
            va="top",
        )

        # Add the selection highlight
        self._selecetd_point = None  # Should use properties
        self._selected_loc = self._location_ax.plot(
            [], [], "o", ms=12, alpha=0.4, color="yellow", visible=False, zorder=5
        )[0]

        # Set up the plot containers
        self.station_figs = []
        unique_rx = np.unique(
            np.array(
                [
                    rx.__class__
                    for src in self._data.survey.source_list
                    for rx in src.receiver_list
                ]
            )
        )
        if rx.PointNaturalSource in unique_rx:
            self.station_figs.append(ApparentResPhsStationPlot())
        if rx.Point3DTipper in unique_rx:
            self.station_figs.append(TipperAmplitudeStationPlot())

        self.freqency_figs = []

    @property
    def dict_comp(self):
        return self._dict_comp

    @dict_comp.setter
    def dict_comp(self, value):
        if self._data_dict is None:
            raise Exception("No data dictionary connected to the object.")
        else:
            if value in self._data_dict.iterkeys():
                self._dict_comp = value
            else:
                raise Exception(
                    "{} is not a key in the connected dictionary.".format(value)
                )

    def view(self):
        """
        Excute the viewer and show the matplotlib window
        """
        self.location_fig.show()
        plt.draw()
        plt.show()

    def _onClick(self, event):
        """
        The click function that is executed when a location in
        plot is clicked the left-mouse button.
        """

        # Get the event artist
        artist = event.artist
        ind = event.ind
        # Get the location
        x, y = artist.get_xdata()[ind][0], artist.get_ydata()[ind][0]
        self._selecetd_point = np.array([[x, y]])
        # Check if the figure exists

        if self._need_setup:
            self._setup_station_figs()
            self._need_setup = False
        else:
            self._clear_station_figs()
        # Draw
        draw_list = [self._data]
        if self._data_dict is not None:
            draw_list.append(self._data_dict[self.dict_comp])
        self._draw_station_figs(draw_list, self._selecetd_point)

        # Highlight the point
        self._selected_loc.set_visible(True)
        self._selected_loc.set_data(x, y)

        # Set the text
        self.text.set_text("Selected: {:.1f} - {:.1f}".format(x, y))
        self.location_fig.canvas.draw()

        for sta_fig in self.station_figs:
            sta_fig.fig.canvas.draw()
            sta_fig.fig.show()

    def _setup_station_figs(self):
        """
        Setup a station data plot figures


        """
        if len(self.station_figs) > 0:
            for sta_fig in self.station_figs:
                sta_fig.setup()

    def _draw_station_figs(self, data_list, location):
        """
        Draw the station figure

        """

        if len(self.station_figs) > 0:
            for sta_fig in self.station_figs:
                sta_fig.draw(data_list, location)

    def _clear_station_figs(self):
        """
        Clear all the axes of the station figure
        """

        if len(self.station_figs) > 0:
            for sta_fig in self.station_figs:
                sta_fig.clear_axes()



"""
===================================================================
Electromagnetics Utilities (:mod:`SimPEG.electromagnetics.utils`)
===================================================================

.. currentmodule:: SimPEG.electromagnetics.utils


Current Utilities
=================

.. autosummary::
  :toctree: generated/

  edge_basis_function
  getStraightLineCurrentIntegral
  segmented_line_current_source_term
  line_through_faces
  getSourceTermLineCurrentPolygon

Waveform Utilities
==================

.. autosummary::
  :toctree: generated/

  omega
  k
  VTEMFun
  convolve_with_waveform

"""
from .waveform_utils import (
    omega,
    k,
    VTEMFun,
    convolve_with_waveform,
)

from .current_utils import (
    edge_basis_function,
    getStraightLineCurrentIntegral,
    getSourceTermLineCurrentPolygon,
    segmented_line_current_source_term,
    line_through_faces,
)



""" module SimPEG.EM.NSEM.Utils

Collection of utilities that are usefull for the NSEM problem

NOTE: These utilities are not well test, use with care

"""
from .solutions_1d import get1DEfields  # Add the names of the functions
from .analytic_1d import getEHfields, getImpedance
from .data_utils import (
    appResPhs,
    rec_to_ndarr,
    rotate_data,
    skindepth,
    makeAnalyticSolution,
    plotMT1DModelData,
    plotImpAppRes,
    printTime,
    convert3Dto1Dobject,
    resample_data,
    extract_data_info,
)
from .edi_files_utils import EDIimporter, _findLatLong, _findLine, _findEDIcomp
from .test_utils import (
    getAppResPhs,
    setup1DSurvey,
    setupSimpegNSEM_ePrimSec,
    random,
    halfSpace,
    blockInhalfSpace,
    twoLayer,
)



from ... import survey
from ...utils import validate_string


class Point(survey.BaseRx):
    """Magnetic point receiver class for integral formulation

    Parameters
    ----------
    locations : (n, 3) numpy.ndarray
        Receiver locations.
    components : str or list of str, default: 'tmi'
        Use a ``str`` for a single component or a ``list`` of ``str`` if multiple
        components are simulated at each location. Component choices are:

        - "tmi"  --> total magnetic intensity data (DEFAULT)
        - "bx"   --> x-component of the magnetic field
        - "by"   --> y-component of the magnetic field
        - "bz"   --> z-component of the magnetic field
        - "bxx"  --> x-derivative of the x-component
        - "bxy"  --> y-derivative of the x-component (and visa versa)
        - "bxz"  --> z-derivative of the x-component (and visa versa)
        - "byy"  --> y-derivative of the y-component
        - "byz"  --> z-derivative of the y-component (and visa versa)
        - "bzz"  --> z-derivative of the z-component

    Notes
    -----
    If predicting amplitude data, you must set include 'bx', 'by', and 'bz' here, and
    set `is_amplitude_data` in the `magnetics.Simulation3DIntegral` to `True`.

    """

    def __init__(self, locations, components="tmi", **kwargs):
        super().__init__(locations, **kwargs)

        if isinstance(components, str):
            components = [components]
        for component in components:
            validate_string(
                "component",
                component,
                [
                    "bxx",
                    "bxy",
                    "bxz",
                    "byy",
                    "byz",
                    "bzz",
                    "bx",
                    "by",
                    "bz",
                    "tmi",
                ],
            )
        self.components = components

    @property
    def nD(self):
        """Number of data.

        Returns
        -------
        int
            Number of data for the receiver (locations X components)
        """
        return self.locations.shape[0] * len(self.components)



import numpy as np

from .optimization import IterationPrinters, StoppingCriteria
from .directives import DirectiveList
from .utils import timeIt, Counter, validate_type, validate_string


class BaseInversion(object):
    """Inversion Class"""

    def __init__(
        self,
        invProb,
        directiveList=None,
        counter=None,
        debug=False,
        name="BaseInversion",
        **kwargs
    ):
        if directiveList is None:
            directiveList = []
        self.directiveList = directiveList
        self.counter = counter
        self.debug = debug
        self.name = name
        super().__init__(**kwargs)

        self.invProb = invProb

        self.opt = invProb.opt
        self.opt.callback = self._optCallback

        self.stoppers = [StoppingCriteria.iteration]

        # Check if we have inserted printers into the optimization
        if IterationPrinters.phi_d not in self.opt.printers:
            self.opt.printers.insert(1, IterationPrinters.beta)
            self.opt.printers.insert(2, IterationPrinters.phi_d)
            self.opt.printers.insert(3, IterationPrinters.phi_m)

    @property
    def name(self):
        """The name of the inversion.

        Returns
        -------
        str
        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = validate_string("name", value)

    #: Print debugging information
    @property
    def debug(self):
        """Debugging flag.

        Returns
        -------
        bool
        """
        return self._debug

    @debug.setter
    def debug(self, value):
        self._debug = validate_type("debug", value, bool)

    #: Set this to a SimPEG.utils.Counter() if you want to count things
    @property
    def counter(self):
        """The counter.

        Returns
        -------
        None or SimPEG.utils.Counter
        """
        return self._counter

    @counter.setter
    def counter(self, value):
        if value is not None:
            value = validate_type("counter", value, Counter, cast=False)
        self._counter = value

    @property
    def directiveList(self):
        if getattr(self, "_directiveList", None) is None:
            self._directiveList = DirectiveList(inversion=self)
            self._directiveList.validate()  # validate if we skip setter
        return self._directiveList

    @directiveList.setter
    def directiveList(self, value):
        if isinstance(value, list):
            value = DirectiveList(*value)
        assert isinstance(value, DirectiveList), "Must be a DirectiveList"
        value.validate()  # validate before setting
        self._directiveList = value
        self._directiveList.inversion = self

    @timeIt
    def run(self, m0):
        """run(m0)

        Runs the inversion!

        """
        self.invProb.startup(m0)
        self.directiveList.call("initialize")
        print("model has any nan: {:b}".format(np.any(np.isnan(self.invProb.model))))
        self.m = self.opt.minimize(self.invProb.evalFunction, self.invProb.model)
        self.directiveList.call("finish")

        return self.m

    def _optCallback(self, xt):
        self.directiveList.call("endIter")



try:
    import SimPEG.dask.simulation
    import SimPEG.dask.electromagnetics.frequency_domain.simulation
    import SimPEG.dask.electromagnetics.static.resistivity.simulation
    import SimPEG.dask.electromagnetics.static.resistivity.simulation_2d
    import SimPEG.dask.electromagnetics.static.induced_polarization.simulation
    import SimPEG.dask.electromagnetics.static.induced_polarization.simulation_2d
    import SimPEG.dask.electromagnetics.time_domain.simulation
    import SimPEG.dask.potential_fields.base
    import SimPEG.dask.potential_fields.gravity.simulation
    import SimPEG.dask.potential_fields.magnetics.simulation
    import SimPEG.dask.simulation
    import SimPEG.dask.data_misfit
    import SimPEG.dask.inverse_problem
    import SimPEG.dask.objective_function

except ImportError as err:
    print("unable to load dask operations")
    print(err)



from ..base_1d import BaseEM1DSimulation
from .sources import StepOffWaveform
from .receivers import (
    PointMagneticFluxDensity,
    PointMagneticField,
    PointMagneticFluxTimeDerivative,
)
import numpy as np

from .survey import Survey
from scipy.constants import mu_0
from scipy.interpolate import InterpolatedUnivariateSpline as iuSpline
from scipy.special import roots_legendre

from empymod import filters
from empymod.transform import get_dlf_points

from geoana.kernels.tranverse_electric_reflections import rTE_forward, rTE_gradient

from ...utils import validate_type, validate_string


class Simulation1DLayered(BaseEM1DSimulation):
    """
    Simulation class for simulating the TEM response over a 1D layered Earth
    for a single sounding.
    """

    def __init__(self, survey=None, time_filter="key_81_CosSin_2009", **kwargs):
        super().__init__(survey=survey, **kwargs)
        self._coefficients_set = False
        self.time_filter = time_filter

    @property
    def survey(self):
        """The survey for the simulation
        Returns
        -------
        SimPEG.electromagnetics.time_domain.survey.Survey
        """
        if self._survey is None:
            raise AttributeError("Simulation must have a survey set")
        return self._survey

    @survey.setter
    def survey(self, value):
        if value is not None:
            value = validate_type("survey", value, Survey, cast=False)
        self._survey = value

    @property
    def time_filter(self):
        return self._time_filter

    @time_filter.setter
    def time_filter(self, value):
        self._time_filter = validate_string(
            "time_filter",
            value,
            ["key_81_CosSin_2009", "key_201_CosSin_2012", "key_601_CosSin_2009"],
        )

        if self._time_filter == "key_81_CosSin_2009":
            self._fftfilt = filters.key_81_CosSin_2009()
        elif self._time_filter == "key_201_CosSin_2012":
            self._fftfilt = filters.key_201_CosSin_2012()
        elif self._time_filter == "key_601_CosSin_2009":
            self._fftfilt = filters.key_601_CosSin_2009()

    def get_coefficients(self):
        if self._coefficients_set is False:
            self._compute_coefficients()
        return (
            self._As,
            self._frequencies,
            self._lambs,
            self._unique_lambs,
            self._inv_lambs,
            self._C0s,
            self._C1s,
        )

    def _set_coefficients(self, coefficients):
        self._As = coefficients[0]
        self._frequencies = coefficients[1]
        self._lambs = coefficients[2]
        self._unique_lambs = coefficients[3]
        self._inv_lambs = coefficients[4]
        self._C0s = coefficients[5]
        self._C1s = coefficients[6]
        self._coefficients_set = True
        return

    def _compute_coefficients(self):
        if self._coefficients_set:
            return
        self._compute_hankel_coefficients()
        survey = self.survey

        t_min = np.infty
        t_max = -np.infty
        x, w = roots_legendre(251)
        # loop through source and receiver lists to find the minimum and maximum
        # evaluation times for the step response
        for src in survey.source_list:
            for rx in src.receiver_list:
                wave = src.waveform
                if isinstance(wave, StepOffWaveform):
                    times = rx.times[rx.times > 0]
                    t_min = min(times.min(), t_min)
                    t_max = max(times.max(), t_max)
                else:
                    try:
                        times = rx.times - wave.time_nodes[:, None]
                        times[times < 0.0] = 0.0
                        quad_points = (times[:-1] - times[1:])[..., None] * (
                            x + 1
                        ) + times[1:, :, None]
                        t_min = min(quad_points[quad_points > 0].min(), t_min)
                        t_max = max(quad_points[quad_points > 0].max(), t_max)
                    except AttributeError:
                        raise TypeError(
                            f"Unsupported source waveform object of {src.waveform}"
                        )

        omegas, t_spline_points = get_dlf_points(self._fftfilt, np.r_[t_min, t_max], -1)
        omegas = omegas.reshape(-1)
        n_omega = len(omegas)
        n_t = len(t_spline_points)

        n_base = len(self._fftfilt.base)
        A_dft = np.zeros((n_t, n_omega))
        for i in range(n_t):
            A_dft[i, i : i + n_base] = self._fftfilt.cos * (-2.0 / np.pi)
        A_dft = A_dft[::-1]  # shuffle these back

        # Calculate the interpolating spline basis functions for each spline point
        splines = []
        for i in range(n_t):
            e = np.zeros(n_t)
            e[i] = 1.0
            sp = iuSpline(np.log(t_spline_points[::-1]), e, k=5)
            splines.append(sp)
        # As will go from frequency to time domain
        As = []
        for src in survey.source_list:
            for rx in src.receiver_list:
                #######
                # Fourier Transform coefficients
                ######
                wave = src.waveform

                def func(t, i):
                    out = np.zeros_like(t)
                    t = t.copy()
                    t[
                        (t > 0.0) & (t <= t_spline_points.min())
                    ] = t_spline_points.min()  # constant at very low ts
                    out[t > 0.0] = splines[i](np.log(t[t > 0.0])) / t[t > 0.0]
                    return out

                # Then calculate the values at each time
                A = np.zeros((len(rx.times), n_t))
                if isinstance(wave, StepOffWaveform):
                    # do not need to do too much fancy here, just need to interpolate
                    # from t_spline_points to rx.times (at positive times)...
                    for i in range(n_t):
                        A[:, i] = func(rx.times, i)
                else:
                    # loop over pairs of nodes and use gaussian quadrature to integrate

                    time_nodes = wave.time_nodes
                    n_interval = len(time_nodes) - 1
                    quad_times = []
                    for i in range(n_interval):
                        b = rx.times - time_nodes[i]
                        b = np.maximum(b, 0.0)
                        a = rx.times - time_nodes[i + 1]
                        a = np.maximum(a, 0.0)
                        quad_times = (b - a)[:, None] * (x + 1) / 2.0 + a[:, None]
                        quad_scale = (b - a) / 2
                        wave_eval = wave.eval_deriv(rx.times[:, None] - quad_times)
                        for i in range(n_t):
                            A[:, i] -= np.sum(
                                quad_scale[:, None]
                                * w
                                * wave_eval
                                * func(quad_times, i),
                                axis=-1,
                            )
                if isinstance(rx, (PointMagneticFluxDensity, PointMagneticField)):
                    As.append(A @ (A_dft / omegas))
                else:
                    As.append(A @ A_dft)
                if isinstance(
                    rx, (PointMagneticFluxTimeDerivative, PointMagneticFluxDensity)
                ):
                    As[-1] *= mu_0

            self._frequencies = omegas / (2 * np.pi)
            self._As = As
        self._coefficients_set = True

    def dpred(self, m, f=None):
        """
        Return predicted data.
        Predicted data, (`_pred`) are computed when
        self.fields is called.
        """
        if f is None:
            f = self.fields(m)

        return f

    def fields(self, m):
        """
        This method evaluates the Hankel transform for each source and
        receiver and outputs it as a list. Used for computing response
        or sensitivities.
        """
        self._compute_coefficients()

        self.model = m

        C0s = self._C0s
        C1s = self._C1s

        frequencies = self._frequencies
        unique_lambs = self._unique_lambs
        inv_lambs = self._inv_lambs
        W = self._W

        sig = self.compute_complex_sigma(frequencies)
        mu = self.compute_complex_mu(frequencies)

        rTE = rTE_forward(frequencies, unique_lambs, sig, mu, self.thicknesses)
        rTE = rTE[:, inv_lambs]
        v = ((C0s * rTE) @ self._fhtfilt.j0 + (C1s * rTE) @ self._fhtfilt.j1) @ W.T

        return self._project_to_data(v.T)

    def getJ(self, m, f=None):
        self.model = m
        if getattr(self, "_J", None) is None:
            self._J = {}
            self._compute_coefficients()

            C0s = self._C0s
            C1s = self._C1s
            lambs = self._lambs
            frequencies = self._frequencies
            unique_lambs = self._unique_lambs
            inv_lambs = self._inv_lambs
            # Had to convert this to an array for a tensor operation (?)
            W = self._W.toarray()

            sig = self.compute_complex_sigma(frequencies)
            mu = self.compute_complex_mu(frequencies)

            if self.hMap is not None:
                # Grab a copy
                C0s_dh = C0s.copy()
                C1s_dh = C1s.copy()
                h_vec = self.h
                i = 0
                for i_src, src in enumerate(self.survey.source_list):
                    h = h_vec[i_src]
                    nD = sum(rx.locations.shape[0] for rx in src.receiver_list)
                    ip1 = i + nD
                    v = np.exp(-lambs[i:ip1] * h)
                    C0s_dh[i:ip1] *= v * -lambs[i:ip1]
                    C1s_dh[i:ip1] *= v * -lambs[i:ip1]
                    i = ip1
                    # J will be n_d * n_src (each source has it's own h)...

                rTE = rTE_forward(frequencies, unique_lambs, sig, mu, self.thicknesses)
                rTE = rTE[:, inv_lambs]
                v_dh_temp = (
                    W
                    @ (
                        (C0s_dh * rTE) @ self._fhtfilt.j0
                        + (C1s_dh * rTE) @ self._fhtfilt.j1
                    ).T
                )
                # need to re-arange v_dh as it's currently (n_data x n_freqs)
                # however it already contains all the relevant information...
                # just need to map it from the rx index to the source index associated..
                v_dh = np.zeros((self.survey.nSrc, *v_dh_temp.shape))

                i = 0
                for i_src, src in enumerate(self.survey.source_list):
                    nD = sum(rx.locations.shape[0] for rx in src.receiver_list)
                    ip1 = i + nD
                    v_dh[i_src, i:ip1] = v_dh_temp[i:ip1]
                    i = ip1
                v_dh = np.transpose(v_dh, (1, 2, 0))
                self._J["dh"] = self._project_to_data(v_dh)

            if (
                self.sigmaMap is not None
                or self.muMap is not None
                or self.thicknessesMap is not None
            ):
                rTE_ds, rTE_dh, rTE_dmu = rTE_gradient(
                    frequencies, unique_lambs, sig, mu, self.thicknesses
                )
                if self.sigmaMap is not None:
                    rTE_ds = rTE_ds[..., inv_lambs]
                    v_ds = (
                        (
                            (C0s * rTE_ds) @ self._fhtfilt.j0
                            + (C1s * rTE_ds) @ self._fhtfilt.j1
                        )
                        @ W.T
                    ).T
                    self._J["ds"] = self._project_to_data(v_ds)
                if self.muMap is not None:
                    rTE_dmu = rTE_dmu[..., inv_lambs]
                    v_dmu = (
                        (
                            (C0s * rTE_dmu) @ self._fhtfilt.j0
                            + (C1s * rTE_dmu) @ self._fhtfilt.j1
                        )
                        @ W.T
                    ).T
                    self._J["dmu"] = self._project_to_data(v_dmu)
                if self.thicknessesMap is not None:
                    rTE_dh = rTE_dh[..., inv_lambs]
                    v_dthick = (
                        (
                            (C0s * rTE_dh) @ self._fhtfilt.j0
                            + (C1s * rTE_dh) @ self._fhtfilt.j1
                        )
                        @ W.T
                    ).T
                    self._J["dthick"] = self._project_to_data(v_dthick)
        return self._J

    def _project_to_data(self, v):
        As = self._As
        if v.ndim == 3:
            out = np.empty((self.survey.nD, v.shape[-1]))
        else:
            out = np.empty((self.survey.nD))
        i_dat = 0
        i_A = 0
        i = 0
        for src in self.survey.source_list:
            for rx in src.receiver_list:
                i_datp1 = i_dat + rx.nD
                n_locs = rx.locations.shape[0]
                i_p1 = i + n_locs
                v_slice = v[np.arange(i, i_p1)]
                # this should order it as location changing faster than time
                # i.e. loc_1 t_1, loc_2 t_1, loc1 t2, loc2 t2
                if v.ndim == 3:
                    if isinstance(rx, (PointMagneticFluxDensity, PointMagneticField)):
                        d = np.einsum("ij,...jk->...ik", As[i_A], v_slice.imag)
                    else:
                        d = np.einsum("ij,...jk->...ik", As[i_A], v_slice.real)
                    out[i_dat:i_datp1] = d.reshape((-1, v.shape[-1]), order="F")
                else:
                    if isinstance(rx, (PointMagneticFluxDensity, PointMagneticField)):
                        d = np.einsum("ij,...j->...i", As[i_A], v_slice.imag)
                    else:
                        d = np.einsum("ij,...j->...i", As[i_A], v_slice.real)
                    out[i_dat:i_datp1] = d.reshape(-1, order="F")
                i_dat = i_datp1
                i = i_p1
                i_A += 1
        return out



import numpy as np

from scipy.constants import mu_0, epsilon_0
from SimPEG.electromagnetics.utils import k


def getKc(freq, sigma, a, b, mu=mu_0, eps=epsilon_0):
    a = float(a)
    b = float(b)
    # return 1./(2*np.pi) * np.sqrt(b / a) * np.exp(-1j*k(freq,sigma,mu,eps)*(b-a))
    return np.sqrt(b / a) * np.exp(-1j * k(freq, sigma, mu, eps) * (b - a))


def _r2(xyz):
    return np.sum(xyz**2, 1)


def _getCasingHertzMagDipole(
    srcloc, obsloc, freq, sigma, a, b, mu=(mu_0, mu_0, mu_0), eps=epsilon_0, moment=1.0
):
    mu = np.asarray(mu)
    Kc1 = getKc(freq, sigma[1], a, b, mu[1], eps)

    nobs = obsloc.shape[0]
    dxyz = obsloc - np.c_[np.ones(nobs)] * np.r_[srcloc]

    r2 = _r2(dxyz[:, :2])
    sqrtr2z2 = np.sqrt(r2 + dxyz[:, 2] ** 2)
    k2 = k(freq, sigma[2], mu[2], eps)

    return Kc1 * moment / (4.0 * np.pi) * np.exp(-1j * k2 * sqrtr2z2) / sqrtr2z2


def _getCasingHertzMagDipoleDeriv_r(
    srcloc, obsloc, freq, sigma, a, b, mu=(mu_0, mu_0, mu_0), eps=epsilon_0, moment=1.0
):
    mu = np.asarray(mu)
    HertzZ = _getCasingHertzMagDipole(
        srcloc, obsloc, freq, sigma, a, b, mu, eps, moment
    )

    nobs = obsloc.shape[0]
    dxyz = obsloc - np.c_[np.ones(nobs)] * np.r_[srcloc]

    r2 = _r2(dxyz[:, :2])
    sqrtr2z2 = np.sqrt(r2 + dxyz[:, 2] ** 2)
    k2 = k(freq, sigma[2], mu[2], eps)

    return -HertzZ * np.sqrt(r2) / sqrtr2z2 * (1j * k2 + 1.0 / sqrtr2z2)


def _getCasingHertzMagDipoleDeriv_z(
    srcloc, obsloc, freq, sigma, a, b, mu=(mu_0, mu_0, mu_0), eps=epsilon_0, moment=1.0
):
    mu = np.asarray(mu)
    HertzZ = _getCasingHertzMagDipole(
        srcloc, obsloc, freq, sigma, a, b, mu, eps, moment
    )

    nobs = obsloc.shape[0]
    dxyz = obsloc - np.c_[np.ones(nobs)] * np.r_[srcloc]

    r2z2 = _r2(dxyz)
    sqrtr2z2 = np.sqrt(r2z2)
    k2 = k(freq, sigma[2], mu[2], eps)

    return -HertzZ * dxyz[:, 2] / sqrtr2z2 * (1j * k2 + 1.0 / sqrtr2z2)


def _getCasingHertzMagDipole2Deriv_z_r(
    srcloc, obsloc, freq, sigma, a, b, mu=(mu_0, mu_0, mu_0), eps=epsilon_0, moment=1.0
):
    mu = np.asarray(mu)
    HertzZ = _getCasingHertzMagDipole(
        srcloc, obsloc, freq, sigma, a, b, mu, eps, moment
    )
    dHertzZdr = _getCasingHertzMagDipoleDeriv_r(
        srcloc, obsloc, freq, sigma, a, b, mu, eps, moment
    )

    nobs = obsloc.shape[0]
    dxyz = obsloc - np.c_[np.ones(nobs)] * np.r_[srcloc]

    r2 = _r2(dxyz[:, :2])
    r = np.sqrt(r2)
    z = dxyz[:, 2]
    sqrtr2z2 = np.sqrt(r2 + z**2)
    k2 = k(freq, sigma[2], mu[2], eps)

    return dHertzZdr * (-z / sqrtr2z2) * (1j * k2 + 1.0 / sqrtr2z2) + HertzZ * (
        z * r / sqrtr2z2**3
    ) * (1j * k2 + 2.0 / sqrtr2z2)


def _getCasingHertzMagDipole2Deriv_z_z(
    srcloc, obsloc, freq, sigma, a, b, mu=(mu_0, mu_0, mu_0), eps=epsilon_0, moment=1.0
):
    mu = np.asarray(mu)
    HertzZ = _getCasingHertzMagDipole(
        srcloc, obsloc, freq, sigma, a, b, mu, eps, moment
    )
    dHertzZdz = _getCasingHertzMagDipoleDeriv_z(
        srcloc, obsloc, freq, sigma, a, b, mu, eps, moment
    )

    nobs = obsloc.shape[0]
    dxyz = obsloc - np.c_[np.ones(nobs)] * np.r_[srcloc]

    r2 = _r2(dxyz[:, :2])
    z = dxyz[:, 2]
    sqrtr2z2 = np.sqrt(r2 + z**2)
    k2 = k(freq, sigma[2], mu[2], eps)

    return (dHertzZdz * z + HertzZ) / sqrtr2z2 * (
        -1j * k2 - 1.0 / sqrtr2z2
    ) + HertzZ * z / sqrtr2z2**3 * (1j * k2 * z + 2.0 * z / sqrtr2z2)


def getCasingEphiMagDipole(
    srcloc, obsloc, freq, sigma, a, b, mu=(mu_0, mu_0, mu_0), eps=epsilon_0, moment=1.0
):
    mu = np.asarray(mu)
    return (
        1j
        * omega(freq)
        * mu
        * _getCasingHertzMagDipoleDeriv_r(
            srcloc, obsloc, freq, sigma, a, b, mu, eps, moment
        )
    )


def getCasingHrMagDipole(
    srcloc, obsloc, freq, sigma, a, b, mu=(mu_0, mu_0, mu_0), eps=epsilon_0, moment=1.0
):
    mu = np.asarray(mu)
    return _getCasingHertzMagDipole2Deriv_z_r(
        srcloc, obsloc, freq, sigma, a, b, mu, eps, moment
    )


def getCasingHzMagDipole(
    srcloc, obsloc, freq, sigma, a, b, mu=(mu_0, mu_0, mu_0), eps=epsilon_0, moment=1.0
):
    mu = np.asarray(mu)
    d2HertzZdz2 = _getCasingHertzMagDipole2Deriv_z_z(
        srcloc, obsloc, freq, sigma, a, b, mu, eps, moment
    )
    k2 = k(freq, sigma[2], mu[2], eps)
    HertzZ = _getCasingHertzMagDipole(
        srcloc, obsloc, freq, sigma, a, b, mu, eps, moment
    )
    return d2HertzZdz2 + k2**2 * HertzZ


def getCasingBrMagDipole(
    srcloc, obsloc, freq, sigma, a, b, mu=(mu_0, mu_0, mu_0), eps=epsilon_0, moment=1.0
):
    mu = np.asarray(mu)
    return mu_0 * getCasingHrMagDipole(
        srcloc, obsloc, freq, sigma, a, b, mu, eps, moment
    )


def getCasingBzMagDipole(
    srcloc, obsloc, freq, sigma, a, b, mu=(mu_0, mu_0, mu_0), eps=epsilon_0, moment=1.0
):
    mu = np.asarray(mu)
    return mu_0 * getCasingHzMagDipole(
        srcloc, obsloc, freq, sigma, a, b, mu, eps, moment
    )



import numpy as np
from geoana.em.fdem.base import skin_depth
from geoana.em.tdem import diffusion_distance

from SimPEG import utils


def get_vertical_discretization(n_layer, minimum_dz, geomtric_factor):
    """
    Creates a list of vertical discretizations generate from a geometric series.

    >>> minimum_dz * geomtric_factor ** np.arange(n_layer)

    Parameters
    ----------
    n_layer : int
        The number of discretizations
    minimum_dz : float
        The smallest discretization
    geometric_factor : float
        The expansion factor of the discretizations

    Returns
    (n_layer) : numpy.ndarray
        The cell widths
    """
    hz = minimum_dz * (geomtric_factor) ** np.arange(n_layer)
    print(
        ">> Depth from the surface to the base of the bottom layer is {:.1f}m".format(
            hz[:].sum()
        )
    )
    return hz


def get_vertical_discretization_frequency(
    frequency,
    sigma_background=0.01,
    n_layer=19,
    hz_min=None,
    z_max=None,
    factor_fmax=4,
    factor_fmin=1.0,
):
    """
    Creates a list of recommended vertical discretizations.

    The vertical discretizations are determined based on the background conductivity
    and the frequency. They are intended to be used in a Layered one dimensional
    simulation.

    Parameters
    ----------
    frequency : numpy.ndarray
        The frequencies needed to represent.
    sigma_background : float, optional
        The background conductivity
    n_layer : int, optional
        Number of layers to generate
    hz_min : optional
        The minimum cell size. By default, it is estimated to be the maximum
        frequency skin depth divided by `factor_fmax`.
    z_max : optional
        The maximum depth of the cells. By default it is estimated as the minimum
        frequency skin depth times `factor_fmin`.
    factor_fmax, factor_fmin : float, optional
        The scaling factors to scale the minimum and maximum skin depth
        estimate, respectively

    Returns
    -------
    (n_layer) numpy.ndarray
        The cell widths.
    """
    if hz_min is None:
        hz_min = skin_depth(frequency.max(), sigma_background) / factor_fmax
    if z_max is None:
        z_max = skin_depth(frequency.min(), sigma_background) * factor_fmin
    i = 4
    hz = np.logspace(np.log10(hz_min), np.log10(hz_min * i), n_layer)
    z_sum = hz.sum()

    while z_sum < z_max:
        i += 1
        hz = np.logspace(np.log10(hz_min), np.log10(hz_min * i), n_layer)
        z_sum = hz.sum()
    return hz


def get_vertical_discretization_time(
    time,
    sigma_background=0.01,
    n_layer=19,
    hz_min=None,
    z_max=None,
    factor_tmin=4,
    facter_tmax=1.0,
):
    """
    Creates a list of recommended vertical discretizations.

    The vertical discretizations are determined based on the background conductivity
    and the times. They are intended to be used in a Layered one dimensional
    simulation.

    Parameters
    ----------
    time : numpy.ndarray
        The times in seconds needed to represent.
    sigma_background : float, optional
        The background conductivity in S/m.
    n_layer : int, optional
        Number of layers to generate
    hz_min : optional
        The minimum cell size in meters. By default, it is estimated to be the minimum diffusion
        distance divided by `factor_tmin`.
    z_max : optional
        The maximum depth of the cells in meters. By default it is estimated as the maximum
        diffusion distance times `factor_tmax`.
    factor_tmin, factor_tmax : float, optional
        The scaling factors to scale the minimum and maximum diffusion distances
        estimate.

    Returns
    -------
    (n_layer) numpy.ndarray
        The cell widths.
    """
    if hz_min is None:
        hz_min = diffusion_distance(time.min(), sigma_background) / factor_tmin
    if z_max is None:
        z_max = diffusion_distance(time.max(), sigma_background) * facter_tmax
    i = 4
    hz = np.logspace(np.log10(hz_min), np.log10(hz_min * i), n_layer)
    z_sum = hz.sum()
    while z_sum < z_max:
        i += 1
        hz = np.logspace(np.log10(hz_min), np.log10(hz_min * i), n_layer)
        z_sum = hz.sum()
    return hz


#############################################################
#       PHYSICAL PROPERTIES
#############################################################


def ColeCole(f, sig_inf=1e-2, eta=0.1, tau=0.1, c=1):
    r"""
    Computing Cole-Cole model in frequency domain

    Parameters
    ----------
    f : float or (n_freq) numpy.ndarray
        Frequency in Hz.
    sig_inf : float or (n_sig) numpy.ndarray, optional
        Conductivity at infinite time.
        If `numpy.ndarray` `f` must also be a `numpy.ndarray`
    eta, tau, c : float, optional
        Cole Cole control parameters

    Returns
    -------
    float, (n_freq) numpy.ndarray, or (n_freq, n_sig) numpy.ndarray

    Notes
    -----
    Defined as

    .. math ::

        \sigma (\omega ) = \sigma_{\infty} \Bigg [
        1 - \eta \Bigg ( \frac{1}{1 + (1-\eta ) (1 + i\omega \tau)^c} \Bigg )
        \Bigg ]

    """

    if np.isscalar(sig_inf):
        w = 2 * np.pi * f
        sigma = sig_inf - sig_inf * eta / (1 + (1 - eta) * (1j * w * tau) ** c)
    else:
        sigma = np.zeros((f.size, sig_inf.size), dtype=complex)
        for i in range(f.size):
            w = 2 * np.pi * f[i]
            sigma[i, :] = utils.mkvc(
                sig_inf - sig_inf * eta / (1 + (1 - eta) * (1j * w * tau) ** c)
            )
    return sigma


def LogUniform(f, chi_inf=0.05, del_chi=0.05, tau1=1e-5, tau2=1e-2):
    r"""
    Computing relaxation model in the frequency domain for a log-uniform
    distribution of time-relaxation constants.

    Parameters
    ----------
    f : float or numpy.ndarray
        Frequency in Hz.
    chi_inf, del_chi, tau1, tau2, optional
        Relaxation model control parameters.

    Returns
    -------
    float or numpy.ndarray

    Notes
    -----

    .. math::

        \chi (\omega ) = \chi_{\infty} + \Delta \chi \Bigg [
        1 - \Bigg ( \frac{1}{ln (\tau_2 / \tau_1 )} \Bigg )
        ln \Bigg ( \frac{1 + i\omega \tau_2}{1 + i\omega tau_1} ) \Bigg )
        \Bigg ]
    """

    w = 2 * np.pi * f
    return chi_inf + del_chi * (
        1 - np.log((1 + 1j * w * tau2) / (1 + 1j * w * tau1)) / np.log(tau2 / tau1)
    )
